package HTML::Mason::Forms;

use strict;
use vars qw($VERSION @ISA @EXPORT @EXPORT_OK);
use Data::Dumper;
use URI;
use Carp;

require Exporter;

@ISA = qw(Exporter);
@EXPORT = qw(
	preprocess
);
$VERSION = '0.03';

sub DEBUG () {0}

#########################################################
use  vars qw($session $session_load $session_save $use_session $err_missing $err_bad
             %dispatch $steps);
%dispatch=(
'MASON'=>\&mason,
'FORM'=>\&form,
'/FORM'=>\&form_END,
'INPUT'=>\&input,
'SELECT'=>\&__select,
'/SELECT'=>\&select_END,
'TEXTAREA'=>\&textarea,
'OPTION'=>\&option,
'NEXT'=>\&__next,
'BYPASS'=>\&bypass,
);

#########################################################
sub preprocess
{
    my($sref)=@_;

    $session=0;
    $use_session=0;
    $steps=0;
    $session_load='';
    $session_save='';
    $err_missing="Missing field(s): ";
    $err_bad="Bad field(s): ";

    local $Data::Dumper::Purity=1;
    _clear_box();
    _clear_select();
    my $change=1 if $$sref=~s(<  
                        (?: (?:%(MASON|INPUT|SELECT|OPTION|FORM|NEXT|BYPASS)) 
                        |
                        (?:/(%SELECT|%FORM))
                    )
                    (.*?)
                 >)
                (_preprocess(uc($1||$2), $3))gisex;

    $change=1 if $$sref=~s(<
                    %(TEXTAREA)(.*?)>(.*?)</%TEXTAREA>
                )
                (_preprocess(uc($1), $2, $3))gisex;

    if($change) {
        select_END();         # in case the users wasn't paying attention...
        finish_boxes();       # close up all check and radio boxes
        $$sref=join '', _perl_init(), 
                        _submit(), 
                        _session_var(),
                        _validation(), 
                        $$sref, 
                        _tailing();
        return 1;
    }
    return 0;
}

#########################################################
sub _preprocess
{
    my($tag, $atts, $extra)=@_;
    $tag=~s(^%)(/);

    my %atts=parse_atts($atts);
    my $ret="<$tag$atts>";

    $ret=$dispatch{$tag}->(\%atts, $ret, $extra) if exists $dispatch{$tag};

    return $ret;
}


#########################################################
sub _mk_first
{
    my($step)=@_;
    return ($steps and $step) ? qq((\$ARGS{'$steps'}||0)!=$step)
                              : qq(keys %ARGS);
}

#########################################################
sub _mk_to_hashref
{
    my($V)=@_;
    return qq(!ref($V) ? {$V => 1} : ref($V) eq 'HASH' ? $V : ref($V) eq 'ARRAY' ? {map {\$_=>1} \@{$V}} : die "\\$V is a ".ref($V));
}

#########################################################
sub _hash_args
{
    my($name, $value, $step)=@_;
    carp "_hash_args must be passed a hashref" unless 'HASH' eq ref $value;
    my $V=qq({});

    my $ses=qq(\$SESSION{'$name'});
    my $S=_mk_to_hashref($ses);

    if($value) {
        $V=Data::Dumper->Dump([$value],[]);
        $V=~s/\s+/ /g;
        $V=~s/\$VAR\d+ = //;
        $V=~s/;\s*$//;
    }


    my $first=_mk_first($step);
    my $arg=qq(\$ARGS{$name});
    my $A=_mk_to_hashref($arg);
    my $S_V=$V;
    $S_V=qq(exists $ses ? ($S) :$V) if $use_session;
    _perl_init(qq(my \%$name=%{exists $arg ? ($A) : $S_V};));
}

#########################################################
sub _mk_to_scalar
{
    my($V)=@_;
    return qq(!ref($V) ? $V : ref($V) eq 'ARRAY' ? $V\->[0] : ref($V) eq 'HASH' ? eval {warn "$V is a HASH!"; return (keys %{$V})[0]} : die "\\$V is a ".ref($V));
}

#########################################################
sub _scalar_args
{
    my($name, $value, $step)=@_;

    my $V=qq("\Q$value\E");
    my $ses=qq(\$SESSION{'$name'});
    my $S=_mk_to_scalar($ses);
    my $arg=qq(\$ARGS{'$name'});
    my $A=_mk_to_scalar($arg);
    my $first=_mk_first($steps);
    my $S_V=$V;
    $S_V=qq(exists $ses ? ($S) :$V) if $use_session;
    _perl_init(qq(my \$$name=exists $arg ? ($A) : $S_V;));
}

#########################################################
{
    my @perl_init;
    sub _perl_init
    {
        if(@_)
        {
            DEBUG and warn "init:", join ', ', grep defined, @_;
            push @perl_init, grep defined, @_;
        } elsif(@perl_init)
        {
            my $t=qq(<%INIT>\n).join "\n", @perl_init, qq(</%INIT>);
            @perl_init=();
            return $t;
        } else
        {
            return '';
        }
    }
}

#########################################################
{
    my @submit;
    sub _submit
    {
        if(@_)
        {
            push @submit, grep defined, @_;
        } elsif(@submit)
        {
            my $t=qq[<%INIT>\nmy \$FIRST=not (].
                  join(" or ", map {qq(exists \$ARGS{$_})} @submit).
                  qq[);\n</%INIT>];
            @submit=();
            return $t;
        } else
        {
            return qq(<%INIT>my \$FIRST=1;\n</%INIT>);
        }
    }
}
#########################################################
{
    my @tailing;
    sub _tailing
    {
        if(@_)
        {
            push @tailing, grep defined, @_;
        } elsif(@tailing)
        {
            my $t=qq[<%INIT>].join("\n", @tailing).qq[\n</%INIT>];
            @tailing=();
            return $t;
        } else
        {
            return "";
        }
    }
}

#########################################################
{
    my @session_vars;
    sub _session_var
    {
        return '' unless $session;
        my $t='';
        my $cleanup='';
        if(@_)
        {
            push @session_vars, grep defined, @_;
        } elsif(@session_vars)
        {
            my($q, $s, %done);
            my $init='';
            foreach my $var (@session_vars)
            {
                next if $done{$var}++;
                $s=substr($var, 0, 1);
                $q=substr($var, 1);

                if($s eq '$'){
                    $init.=qq(    $var = \$SESSION{'$q'}||'' if exists \$SESSION{'$q'};)
                } else {
                    $init.=qq(    $var = $s\{\$SESSION{'$q'}} if exists \$SESSION{'$q'};)
                }
                $init.="\n";
                
                $cleanup.=qq(        \$SESSION{'$q'}=) . ($s eq '$' ? '' : '\\') . qq($var;\n);
            }

            @session_vars=();
            $t=<<MASON if $init;
<%INIT>
$session_load
if(\$FIRST and defined \%SESSION) {
$init
}
</%INIT>
MASON
        }

        $t.=<<MASON;
<%INIT>
my \$SAVE_SESSION=sub {
    if(defined \%SESSION) {
$cleanup
        tied(%SESSION)->save() if tied(%SESSION) and tied(%SESSION)->can('save');
    }
};
</%INIT>
<%cleanup>
&\$SAVE_SESSION if defined \%SESSION;
$session_save
</%cleanup>
MASON
        return $t;
    }
}

#########################################################
{
    my @validation;
    sub _validation
    {
        if(@_)
        {
            push @validation, grep defined, @_;
        } elsif(@validation)
        {
            my $v=join "\n    ", @validation;
            my $t=<<MASON;
<%INIT>
my \$VALIDATION_ERROR='';
unless(\$FIRST) {
    my(\@err, \@missing, \@bad);
    local \$SIG{__DIE__}=sub {push \@err, join '', \@_};
    $v
    push \@err, "$err_bad ".join ', ', \@bad if \@bad;
    push \@err, "$err_missing ".join ', ', \@missing if \@missing;
    \$VALIDATION_ERROR=join '<BR>', \@err if \@err;
}
</%INIT>
MASON
            @validation=();
            return $t;
        } else
        {
            return qq(<%INIT>my \$VALIDATION_ERROR=0;\n</%INIT>);
        }
    }
}

#########################################################
sub mason
{
    my($atts)=@_;
    my $src=$atts->{src};
    unless($src) {
        # warn "<%MASON> tag doesn't have a SRC attribute" unless $src;
        die "<%MASON> tag doesn't have a SRC attribute" unless $src;
    }
    delete $atts->{src};
#    local $Data::Dumper::Useqq=1;
    my $a=Data::Dumper->Dump([$atts], [qw(*my_args)]);
    $a=~s('\$\{(\w+?)\}')(\$$1)g;
    return join '', "<%perl>{my ${a}mc_comp('$src', \%my_args);}</%perl>";
}

#########################################################
{
    my(%checkes, %radios);
    sub _clear_box
    {
        %checkes=();
        %radios=();
    }

    #########################################################
    sub finish_boxes
    {
        local $_;
        my($a, @one);
        # Create variables for the check boxes
        foreach my $name (keys %checkes) {
            # multiple checkboxes becomes a hash
            my %checked=map {@$_} grep {$_->[0]} @{$checkes{$name}};

            # HACK : we assume that all of this checkbox are in this session
            my $s=(grep defined, map {$_->[1]} @{$checkes{$name}})[0];
            _hash_args($name, \%checked, $s);
        }

        # Create variables for the radio buttons
        foreach my $name (keys %radios) {
            next unless ref $radios{$name};
            _scalar_args($name, @{$radios{$name}});
        }
    }

    #########################################################
    sub input
    {
        my($atts, $ret)=@_;
        $atts->{type}||='TEXT';
        my $type=uc $atts->{type};

        unless($atts->{name}) {
            warn qq(<%INPUT type=$type> without a name\n) unless $atts->{name};
            return $ret;
        }
        my $name=_clean_name($atts->{name});
        my $value=$atts->{value}||'';

        my $boxes;
        my $s='$';
        if($type eq 'TEXT' or $type eq 'PASSWORD' or
           $type eq 'HIDDEN') {
            delete @{$atts}{qw(value)};
            _scalar_args($name, $value, $atts->{step});
            $ret=qq(<input value="<%\$$name%>");
        } 
        elsif($type eq 'RADIO') {
            if(exists $atts->{checked}) {
                $radios{$name}||=[$value, $atts->{step}];
            } 
            elsif(not exists $radios{$name}) {
                $radios{$name}='';
            }

            delete @{$atts}{qw(checked)};
            $ret=qq(<input<%(\$$name eq "$value" ? ' checked="1"' : '')%>);
        } 
        elsif($type eq 'CHECKBOX') {
            if(exists $atts->{checked})
            {
                push @{$checkes{$name}}, [$value, $atts->{step}];
            } 
            elsif(not exists $checkes{$name}) {
                $checkes{$name}=[''];
            }
            $s='%';

            delete @{$atts}{qw(checked)};
            $ret=qq(<input<%(exists \$$name\{"\Q$value\E"} ? ' checked="1"' : '')%>);
        } 
        elsif($type eq 'SUBMIT') {
            _submit($name, delete $atts->{step});
            return $ret;
        } 
        else {
            return $ret;
        }
        _std_var($s, $name, $atts);

        delete $atts->{step};
        my $more=_nice_atts($atts);
        return $ret."$more>";
    }
}

#########################################################
{
    my($select, $select_step, %chosen);
    sub _clear_select
    {
        $select='';
        $select_step=0;
        %chosen=();
    }

    #########################################################
    sub __select
    {
        my($atts, $ret)=@_;
        if($select)
        {
            warn qq(Unclosed <%SELECT name="$select">\n);
            $ret=select_END($ret);
        }
        unless($atts->{name}) {
            warn qq(<%SELECT> without a name\n) unless $atts->{name};
            return $ret;
        }
        $select=_clean_name($atts->{name});
        $select_step=$atts->{step};
        %chosen=();
        return $ret;
    }

    #########################################################
    sub select_END
    {
        my($atts, $ret)=@_;

        _hash_args($select, \%chosen, $select_step) if $select;

        _clear_select();
        return $ret;
    }

    #########################################################
    sub option
    {
        my($atts, $ret)=@_;
        my $value=quotemeta $atts->{value};
        my $t=yesno_att($atts, 'selected');

        $chosen{$value}=1 if $t;
        delete $atts->{selected};
        
        my $more=_nice_atts($atts);
        $ret=qq(<OPTION$more<%(exists(\$$select\{"$value"}) ? ' selected="1"' : '')%>>);
        return $ret;
    }

}

#########################################################
sub _std_var
{
    my($s, $name, $atts)=@_;;

    # create the session var
    _session_var("$s$name") unless yesno_att($atts, 'notsession');

    # do required fields
    my $desc=$atts->{desc}||$name;
    $desc=~s/'/\'/g;
    _validation(qq(push \@missing, '$desc' unless $s$name;))
                    if yesno_att($atts, 'required');

    # error message for field validation
    my $err=$atts->{error}||'';
    $err=~s/'/\'/g;
    $err= $err ? qq[push \@error, '$err'] : qq[push \@bad, '$desc'];

    # validation code
    _validation(_nice_code(qq($err unless $atts->{valid};)))
                    if $atts->{'valid'};
    # validation regex
    _validation(qq[$err unless $s$name=~m"$atts->{regex}";])
                    if $atts->{'regex'};

    # get rid of attributes that we used
    delete @{$atts}{qw(required error valid regex notsession desc)};
}

#########################################################
sub textarea
{
    my($atts, $ret, $default)=@_;
    $default=~s/^\s+//;
    $default=~s/\s+$//;

    my $name=_clean_name($atts->{name});
    _scalar_args($name, $default, $atts->{step});

    _std_var('$', $name, $atts);

    delete $atts->{step};
    my $more=_nice_atts($atts);
    return qq(<TEXTAREA$more><%\$$name%></TEXTAREA>);
}

#########################################################
sub form
{
    my($atts, $ret)=@_;
    $session=lc $atts->{session};
    $session=0 if $session eq 'off' or $session eq 'no';
    $session_load='';
    $session_load=_nice_code($atts->{load}) if $atts->{load};
    $session_save='';
    $session_save=_nice_code($atts->{save}) if $atts->{save};

    $steps=$atts->{steps};
    $steps="step" if $steps 
		   and ($steps eq 'on' or $steps eq 'yes' or $steps eq '1');
    $steps=0      if $steps and ($steps eq 'off' or $steps eq 'no');

    $err_bad=$atts->{bad}           if yesno_att($atts, 'bad');
    $err_missing=$atts->{missing}   if yesno_att($atts, 'missing');

    $use_session=yesno_att($atts, 'usesession');

    delete @{$atts}{qw(session load save bad missing steps use_session)};
    my $more=_nice_atts($atts);
    return qq(<%INIT>my \$BYPASS=0;\n</%INIT>\n<FORM$more>);
}

#########################################################
sub form_END
{
    my($atts, $ret)=@_;
    return $ret;
}

#########################################################
sub bypass
{
    my($atts, $ret)=@_;
    return qq(<%INIT>\$BYPASS=1 if($atts->{if});\n</%INIT>)         
                if $atts->{if};
    return qq(<%INIT>\$BYPASS=1 unless($atts->{unless});\n</%INIT>) 
                if $atts->{unless};
    die "<%BYPASS> requires if or unless attributes";
}



#########################################################
sub __next
{
    my($atts, $ret)=@_;

    my $href=$atts->{href};
    my $id=$atts->{session};
    unless($href) {
        DEBUG and warn "<%NEXT> requires an href attribute\n";
        die "<%NEXT> requires an href attribute\n"  unless $href;
        return '';
    }
    $id||='';
    $id=qq(\$u->query_form(\$u->query_form(), eval {return($id)}); die "Error in user code: \$@" if \$@;)
        if $id;
    
    my $sess='';
    $sess='&$SAVE_SESSION;' if $session;

    _tailing(<<MASON);
if((not \$FIRST and not \$VALIDATION_ERROR) or \$BYPASS)
{
    $sess
    my \$u=URI->new_abs("$href", 'http:'.\$r->parsed_uri->path());
    \$u->scheme('');
    $id
#    \$r->status(Apache::Contants::HTTP_REDIRECT());
    \$r->cgi_header_out(Location=>\$u->as_string());
    \$r->send_http_header();
    HTML::Mason::Commands::mc_suppress_http_header(1);
    return;
}
MASON
    return '';
}






#############################################################################
# Helper Functions

#########################################################
sub _clean_name
{
    my($t)=@_;
    $t=~s/\W/_/g;           # turn into a valid identifier
    $t="_$t" unless $t=~/^[_A-Za-z]/;
    return $t;
}

#########################################################
sub _nice_code
{
    my($code)=@_;
    $code.=';' unless $code=~/;\s*$/;
    return qq(eval {$code}; die "Error in user code: \$@\\n" if \$@;);
}

#########################################################
sub _nice_atts
{
    my($atts)=@_;
    return '' unless keys %$atts;
    return join ' ', '', map { defined($atts->{$_}) ? 
                                    qq($_="$atts->{$_}") : $_
                             } keys %$atts;
}

#########################################################
sub yesno_att
{
    my($atts, $key)=@_;
    return unless exists $atts->{$key};         # attribute not present
    return 1 unless defined($atts->{$key});     # <TAG att> equiv =1
    return 1 if $atts->{$key};
    return;
}


#########################################################
### STOLEN FROM HTML::SimpleParse
sub parse_atts             
{
    my($atts)=@_;
    my @returns;
        
    pos($atts) = 0;  # Make sure we start searching at the beginning of the string

    while (1) 
    {
        next if $atts =~ m/\G\s+/gc;  # Get rid of leading whitespace

        if ( $atts =~ m/\G
                (\S+?)=                                  # the key
                (?:
                 "([^\"\\]*  (?: \\.[^\"\\]* )* )"\s*    # quoted string, with possible whitespace inside,
                  |                                      #  or
                 ([^\s>]*)\s*                            # anything else, without whitespace or >
                )/gcx ) 
        {
            my ($key, $val) = ($1, $+);
            $val =~ s/\\(.)/$1/gs;
            push @returns, lc($key), $val;

        } elsif ( $atts =~ m/\G(\S+)\s*/gc ) 
        {
            push @returns, lc($1), undef;
        } else 
        {
            last;
        }
    }

    return @returns;
}




1;
__END__

=head1 NAME

HTML::Mason::Forms - Mason extension for luser-safe component syntax

=head1 SYNOPSIS

    use HTML::Mason::Forms;
    use HTML::Mason;
    my $parser = new HTML::Mason::Parser 
            (
                taint_check=>1, 
                'preprocess'=>\&HTML::Mason::Forms::preprocess,
                ....
            );


=head1 DESCRIPTION

HTML::Mason::Forms is a Mason pre-processor that adds some luser-safe 
HTML-like syntax to Mason components.

=over 4

=item E<lt>%MASON src="header.mason" title="Hello world" root=".."E<gt>

This is a wraper around mc_comp().  The SRC attribute is the component file
name, all other attributes are given as arguements to the component.  The
code above is equivalent to the following.

    %mc_comp("header.mason", title=>"Hello world", root=>"..");

=item E<lt>%INPUT type=input name="e-mail" value="your@email.com">

Short cut for text input.  This will output an E<lt>INPUT> that will take it's
current value from $ARGS, with the VALUE attribute as default. 
Additionnaly, the variable $e_mail is created via the E<lt>%args>

The code above is equivalent to the following:

    <%args>
    ${e-mail}="your@email.com"
    </%args>
    <INPUT type=input name="e-mail" value="<%$e_mail%>">

Note that the name "e-mail" is changed into a valid Perl identifier ($e_mail). 
And that $e_mail isn't HTML escaped.

=item E<lt>%INPUT type=passwd name=E<34>don't showE<34> value=E<34>your passwordE<34>>

Similar short cut as E<lt>INPUT type=text> above

The code above is equivalent to the following:

    <%args>
    ${don't show}="your password"
    </%args>
    <INPUT type=input name="don't show" value="<%$don_t_show%>">

=item E<lt>%INPUT type=checkbox name=meal value=supper>

Short cut for check boxes.  This will output an E<lt>INPUT> that will 
be CHECKED if it was set in $ARGS.  If you want it CHECKED as a default,
included the CHECKED attribute in the tag.
Additionnaly, the variable %meal is created via the E<lt>%args>

An example

    <%INPUT type=checkbox name=meal value="breakfast"> Breakfast<BR>
    <%INPUT type=checkbox name=meal value="lunch" CHECKED> Lunch<BR>
    <%INPUT type=checkbox name=meal value="tea"> Tea <BR>
    <%INPUT type=checkbox name=meal value="supper" CHECKED> Supper<BR>

Is equivalent to

    <%args>
    %meal=>(supper=>1, lunch=>1)
    </%args>
    <INPUT type=checkbox name="meal" value="breakfast" <%exists $meal{breakfast} ? ' CHECKED' : ''%>> Breakfast<BR>
    <INPUT type=checkbox name="meal" value="lunch" <%exists $meal{lunch} ? ' CHECKED' : ''%>> Lunch<BR>
    <INPUT type=checkbox name="meal" value="tea" <%exists $meal{tea} ? ' CHECKED' : ''%>> Tea<BR>
    <INPUT type=checkbox name="meal" value="supper" <%exists $meal{supper} ? ' CHECKED' : ''%>> Supper<BR>

=item E<lt>%INPUT type=radio name=city value="Montr&eacute;al">

Short cut for radio buttons.  This will output an E<lt>INPUT> that will 
be CHECKED if it was set in $ARGS.  If you want it CHECKED as a default,
included the CHECKED attribute in the tag.
If more then one CHECKED tag is present the first one is used.
Additionnaly, the variable $city is created via the E<lt>%args>

An example
    
    <%INPUT type=radio name=city value="mtl" CHECKED> 
    Montr&eacute;<BR>
    <%INPUT type=radio name=city value="qc"> 
    Qu&eacute;bec <BR>
    <%INPUT type=radio name=city value="lvl"> 
    Laval<BR>

Is equivalent to

    <%args>
    $city=>"mtl"
    </%args>
    <INPUT type=radio name=city value="mtl"<%($city||'') eq 'mtl' ? ' CHECKED ' : ''$>> 
    Montr&eacute;<BR>
    <INPUT type=radio name=city value="qc"<%($city||'') eq 'qc' ? ' CHECKED ' : ''$>> 
    Qu&eacute;bec <BR>
    <INPUT type=radio name=city value="lvl"<%($city||'') eq 'lvl' ? ' CHECKED ' : ''$>> 
    Laval<BR>

=head1 AUTHOR

Philip Gwyn, mason@artware.qc.ca

=head1 SEE ALSO

HTML::Mason, HTML::Mason::Parser

=cut

